KOMLOG_USERNAME='username'
KOMLOG_KEYFILE='keyfile'
ENTRY_LOG='logging'
ENTRY_TRANSFERS='transfers'
LOG_FILE='filename'
LOG_DIR='dirname'
LOG_LEVEL='level'
LOG_ROTATION='rotation'
LOG_MAX_BYTES='max_bytes'
LOG_BACKUP_COUNT='backup_count'
LOG_FORMAT='format'
TRANSFERS_ENABLED='enabled'
TRANSFERS_FILE='filename'

